# -*- coding: utf-8 -*-
#
#   TLV: Tag Length Value
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import abstractmethod
from typing import TypeVar, Generic, Union, Optional

from udp.ba import ByteArray, Data
from udp.ba import IntegerData, UInt8Data, UInt16Data, UInt32Data, VarIntData, Convert


"""
       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |         Type                  |            Length             |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         Value (variable)                ....
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
"""


class Tag(ByteArray):
    """ TLV Tag/Type """
    pass


T = TypeVar('T')  # Type


class TagParser(Generic[T]):
    """ Tag Parser """

    @abstractmethod
    def parse_tag(self, data: ByteArray) -> Optional[T]:
        """ Parse Tag from data """
        raise NotImplemented


"""
    Tags
    ~~~~
"""


class Tag8(UInt8Data, Tag):
    """ Fixed size Tag (8 bits) """

    ZERO = None  # Tag8.parse(data=UInt8Data.ZERO)

    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray]):  # -> Tag8
        """ parse Tag """
        if isinstance(data, cls):
            return data
        elif not isinstance(data, UInt8Data):
            data = UInt8Data.from_data(data=data)
        if data is not None:
            return cls(data=data, value=data.value)

    @classmethod
    def new(cls, value: int):  # -> Tag8
        data = UInt8Data.from_int(value=value)
        return cls(data=data, value=data.value)


class Tag16(UInt16Data, Tag):
    """ Fixed size Tag (16 bits) """

    ZERO = None  # Tag16.parse(data=UInt16Data.ZERO)

    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray]):  # -> Tag16
        """ parse Tag """
        if isinstance(data, cls):
            return data
        elif not isinstance(data, UInt16Data):
            data = Convert.uint16data_from_data(data=data)
        if data is not None:
            return cls(data=data, value=data.value, endian=data.endian)

    @classmethod
    def new(cls, value: int):  # -> Tag16
        data = Convert.uint16data_from_value(value=value)
        return cls(data=data, value=data.value, endian=data.endian)


class Tag32(UInt32Data, Tag):
    """ Fixed size Tag (32 bits) """

    ZERO = None  # Tag32.from_data(data=UInt32Data.ZERO)

    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray]):  # -> Tag32
        """ parse Tag """
        if isinstance(data, cls):
            return data
        elif not isinstance(data, UInt32Data):
            data = Convert.uint32data_from_data(data=data)
        if data is not None:
            return cls(data=data, value=data.value, endian=data.endian)

    @classmethod
    def new(cls, value: int):  # -> Tag32
        data = Convert.uint32data_from_value(value=value)
        return cls(data=data, value=data.value, endian=data.endian)


Tag8.ZERO = Tag8.parse(data=UInt8Data.ZERO)
Tag16.ZERO = Tag16.parse(data=UInt16Data.ZERO)
Tag32.ZERO = Tag32.parse(data=UInt32Data.ZERO)


"""
    Variable size Tag
    ~~~~~~~~~~~~~~~~~
    A tag that starts with a variable integer indicating its content length
    
         0                   1                   2                   3
         0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |     Length    |               Content (variable)              |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
"""


class VarTag(Data, Tag):

    ZERO = None  # VarTag.new(content=Data.ZERO)

    def __init__(self, data: Union[bytes, bytearray, ByteArray], length: IntegerData, content: ByteArray):
        if isinstance(data, ByteArray):
            super().__init__(buffer=data.buffer, offset=data.offset, size=data.size)
        else:
            super().__init__(buffer=data)
        self.__length = length
        self.__content = content

    @property
    def length(self) -> IntegerData:
        return self.__length

    @property
    def content(self) -> ByteArray:
        return self.__content

    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray]):  # -> VarTag
        """ parse Tag """
        if isinstance(data, cls):
            return data
        # get length
        length = VarIntData.from_data(data=data)
        if length is None:
            return None
        start = length.size
        end = start + length.value
        if end > data.size:
            # raise ValueError('VarTag length error: %d, %d, %d' % (start, end, data.size))
            return None
        elif end < data.size:
            # cut the tail
            data = data.slice(start=0, end=end)
        # get content
        content = data.slice(start=start, end=end)
        return cls(length=length, content=content, data=data)

    @classmethod
    def new(cls, content: ByteArray, length: Optional[VarIntData] = None):  # -> VarTag
        if length is None:
            length = VarIntData.from_int(value=content.size)
        data = length.concat(content)
        return cls(data=data, length=length, content=content)


VarTag.ZERO = VarTag.new(content=Data.ZERO)



# -*- coding: utf-8 -*-
#
#   TLV: Tag Length Value
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import abstractmethod
from typing import TypeVar, Generic, Union, Optional

from udp.ba import ByteArray
from udp.ba import IntegerData, UInt8Data, UInt16Data, UInt32Data, VarIntData, Convert


from .tag import Tag, T


"""
       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |         Type                  |            Length             |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         Value (variable)                ....
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
"""


class Length(IntegerData):
    """ TLV Length """
    pass


L = TypeVar('L')  # Length


class LengthParser(Generic[T, L]):
    """ Length Parser """

    @abstractmethod
    def parse_length(self, data: ByteArray, tag: T) -> Optional[L]:
        """ Parse Length from data with Tag """
        raise NotImplemented


"""
    Lengths
    ~~~~~~~
"""


class Length8(UInt8Data, Length):
    """ Fixed size Length (8 bits) """

    ZERO = None  # Length8.parse(data=UInt8Data.ZERO)

    # noinspection PyUnusedLocal
    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray], tag: Optional[Tag] = None):  # -> Length8
        """ parse Length """
        if isinstance(data, cls):
            return data
        elif not isinstance(data, UInt8Data):
            data = UInt8Data.from_data(data=data)
        if data is not None:
            return cls(data=data, value=data.value)

    @classmethod
    def new(cls, value: int):  # -> Length8
        data = UInt8Data.from_int(value=value)
        return cls(data=data, value=data.value)


class Length16(UInt16Data, Length):
    """ Fixed size Length (16 bits) """

    ZERO = None  # Length16.parse(data=UInt16Data.ZERO)

    # noinspection PyUnusedLocal
    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray], tag: Optional[Tag] = None):  # -> Length16
        """ parse Length """
        if isinstance(data, cls):
            return data
        elif not isinstance(data, UInt16Data):
            data = Convert.uint16data_from_data(data=data)
        if data is not None:
            return cls(data=data, value=data.value, endian=data.endian)

    @classmethod
    def new(cls, value: int):  # -> Length16
        data = Convert.uint16data_from_value(value=value)
        return cls(data=data, value=data.value, endian=data.endian)


class Length32(UInt16Data, Length):
    """ Fixed size Length (32 bits) """

    ZERO = None  # Length32.parse(data=UInt32Data.ZERO)

    # noinspection PyUnusedLocal
    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray], tag: Optional[Tag] = None):  # -> Length32
        """ parse Length """
        if isinstance(data, cls):
            return data
        elif not isinstance(data, UInt32Data):
            data = Convert.uint32data_from_data(data=data)
        if data is not None:
            return cls(data=data, value=data.value, endian=data.endian)

    @classmethod
    def new(cls, value: int):  # -> Length32
        data = Convert.uint32data_from_value(value=value)
        return cls(data=data, value=data.value, endian=data.endian)


class VarLength(VarIntData, Length):
    """ Variable size Length """

    ZERO = None  # VarLength.parse(data=VarIntData.ZERO)

    # noinspection PyUnusedLocal
    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray], tag: Optional[Tag] = None):  # -> VarLength
        """ parse Length """
        if isinstance(data, cls):
            return data
        elif not isinstance(data, VarIntData):
            data = VarIntData.from_data(data=data)
        if data is not None:
            return cls(data=data, value=data.value)

    @classmethod
    def new(cls, value: int):  # -> VarLength
        data = VarIntData.from_int(value=value)
        return cls(data=data, value=data.value)


Length8.ZERO = Length8.parse(data=UInt8Data.ZERO)
Length16.ZERO = Length16.parse(data=UInt16Data.ZERO)
Length32.ZERO = Length32.parse(data=UInt32Data.ZERO)
VarLength.ZERO = VarLength.parse(data=VarIntData.ZERO)



# -*- coding: utf-8 -*-
#
#   TLV: Tag Length Value
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import abstractmethod
from typing import TypeVar, Generic, Union, Optional

from udp.ba import ByteArray, Data
from udp.ba import UInt8Data, UInt16Data, UInt32Data, Convert


from .tag import Tag, T
from .length import Length, L


"""
       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |         Type                  |            Length             |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         Value (variable)                ....
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
"""


class Value(ByteArray):
    """ TLV Value """
    pass


V = TypeVar('V')  # Value


class ValueParser(Generic[T, L, V]):
    """ Value Parser """

    @abstractmethod
    def parse_value(self, data: ByteArray, tag: T, length: L) -> Optional[V]:
        """ Parse Value from data with Tag & Length """
        raise NotImplemented


"""
    Values
    ~~~~~~
"""


class Value8(UInt8Data, Value):
    """ Fixed size Value (8 bits) """

    ZERO = None  # Value8.parse(data=UInt8Data.ZERO)

    # noinspection PyUnusedLocal
    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray],
              tag: Optional[Tag] = None, length: Optional[Length] = None):  # -> Value8
        """ parse Value """
        if isinstance(data, cls):
            return data
        elif not isinstance(data, UInt8Data):
            data = UInt8Data.from_data(data=data)
        if data is not None:
            return cls(data=data, value=data.value)

    @classmethod
    def new(cls, value: int):  # -> Value8
        data = UInt8Data.from_int(value=value)
        return cls(data=data, value=data.value)


class Value16(UInt16Data, Tag):
    """ Fixed size Value (16 bits) """

    ZERO = None  # Value16.parse(data=UInt16Data.ZERO)

    # noinspection PyUnusedLocal
    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray],
              tag: Optional[Tag] = None, length: Optional[Length] = None):  # -> Value16
        """ parse Value """
        if isinstance(data, cls):
            return data
        elif not isinstance(data, UInt16Data):
            data = Convert.uint16data_from_data(data=data)
        if data is not None:
            return cls(data=data, value=data.value, endian=data.endian)

    @classmethod
    def new(cls, value: int):  # -> Value16
        data = Convert.uint16data_from_value(value=value)
        return cls(data=data, value=data.value, endian=data.endian)


class Value32(UInt32Data, Tag):
    """ Fixed size Value (32 bits) """

    ZERO = None  # Value32.parse(data=UInt32Data.ZERO)

    # noinspection PyUnusedLocal
    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray],
              tag: Optional[Tag] = None, length: Optional[Length] = None):  # -> Value32
        """ parse Value """
        if isinstance(data, cls):
            return data
        elif not isinstance(data, UInt32Data):
            data = Convert.uint32data_from_data(data=data)
        if data is not None:
            return cls(data=data, value=data.value, endian=data.endian)

    @classmethod
    def new(cls, value: int):  # -> Value32
        data = Convert.uint32data_from_value(value=value)
        return cls(data=data, value=data.value, endian=data.endian)


class RawValue(Data, Value):
    """ Base Value """

    ZERO = None  # RawValue.parse(data=Data.ZERO)

    def __init__(self, data: Union[bytes, bytearray, ByteArray]):
        if isinstance(data, ByteArray):
            super().__init__(buffer=data.buffer, offset=data.offset, size=data.size)
        else:
            super().__init__(buffer=data)

    # noinspection PyUnusedLocal
    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray],
              tag: Optional[Tag] = None, length: Optional[Length] = None):  # -> RawValue
        """ parse Value """
        if isinstance(data, cls):
            return data
        else:
            return cls(data=data)

    @classmethod
    def new(cls, data: Union[bytes, bytearray, ByteArray]):  # -> RawValue
        if isinstance(data, cls):
            return data
        else:
            return cls(data=data)


class StringValue(Data, Value):
    """ String Value """

    ZERO = None  # StringValue.parse(data=Data.ZERO)

    def __init__(self, data: Union[bytes, bytearray, ByteArray], string: str):
        if isinstance(data, ByteArray):
            super().__init__(buffer=data.buffer, offset=data.offset, size=data.size)
        else:
            super().__init__(buffer=data)
        self.__string = string

    def __str__(self) -> str:
        return self.__string

    def __repr__(self) -> str:
        return self.__string

    @property
    def string(self) -> str:
        return self.__string

    # noinspection PyUnusedLocal
    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray],
              tag: Optional[Tag] = None, length: Optional[Length] = None):  # -> StringValue
        """ parse Value """
        if isinstance(data, cls):
            return data
        if isinstance(data, ByteArray):
            data = data.get_bytes()
        string = data.decode('utf-8')
        return cls(data=data, string=string)

    @classmethod
    def new(cls, string: str):  # -> StringValue
        data = string.encode('utf-8')
        return cls(data=data, string=string)


Value8.ZERO = Value8.parse(data=UInt8Data.ZERO)
Value16.ZERO = Value16.parse(data=UInt16Data.ZERO)
Value32.ZERO = Value32.parse(data=UInt32Data.ZERO)
RawValue.ZERO = RawValue.parse(data=Data.ZERO)
StringValue.ZERO = StringValue.parse(data=Data.ZERO)



# -*- coding: utf-8 -*-
#
#   TLV: Tag Length Value
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import abstractmethod
from typing import TypeVar, Generic, Optional, List

from udp.ba import ByteArray

from .tag import T
from .length import L
from .value import V


"""
       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |         Type                  |            Length             |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         Value (variable)                ....
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
"""


class Entry(ByteArray, Generic[T, L, V]):
    """ TLV Entry """

    @property
    def tag(self) -> T:
        raise NotImplemented

    @property
    def length(self) -> L:
        raise NotImplemented

    @property
    def value(self) -> V:
        raise NotImplemented


E = TypeVar('E')  # Entry


class EntryParser(Generic[E]):
    """ Entry Parser """

    @abstractmethod
    def parse_entry(self, data: ByteArray) -> Optional[E]:
        """ Parse Entry from data """
        raise NotImplemented

    @abstractmethod
    def parse_entries(self, data: ByteArray) -> List[E]:
        """ Parse all TLV entries """
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   STUN: Session Traversal Utilities for NAT
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from .protocol import *

from .node import *
from .server import *
from .client import *

name = "STUN"

__author__ = 'Albert Moky'

__all__ = [

    #
    #  Protocol
    #
    'MessageType', 'TransactionID', 'Header', 'Package',

    'AttributeType', 'AttributeLength', 'AttributeValue',
    'Attribute', 'AttributeParser',

    'MappedAddressValue',
    'XorMappedAddressValue', 'XorMappedAddressValue2',
    'SourceAddressValue', 'ResponseAddressValue', 'ChangedAddressValue',
    'ChangeRequestValue', 'SoftwareValue',

    #
    #  Service
    #
    'NatType',
    'Client', 'Server',
]



# -*- coding: utf-8 -*-
#
#   STUN: Session Traversal Utilities for NAT
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Session Traversal Utilities for NAT
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    [RFC] https://www.ietf.org/rfc/rfc5389.txt
"""

import threading
from typing import Union, Optional

from udp.ba import ByteArray, Data, MutableData
from udp.ba import Endian, UInt16Data, Convert


"""
                        0                 1
                        2  3  4 5 6 7 8 9 0 1 2 3 4 5

                       +--+--+-+-+-+-+-+-+-+-+-+-+-+-+
                       |M |M |M|M|M|C|M|M|M|C|M|M|M|M|
                       |11|10|9|8|7|1|6|5|4|0|3|2|1|0|
                       +--+--+-+-+-+-+-+-+-+-+-+-+-+-+

                Figure 3: Format of STUN Message Type Field
"""


class MessageType(UInt16Data):

    def __init__(self, data: Union[bytes, bytearray, ByteArray], value: int, endian: Endian, name: str):
        super().__init__(data=data, value=value, endian=endian)
        self.__name = name

    def __str__(self):
        return self.__name

    def __repr__(self):
        return self.__name

    @property
    def name(self) -> str:
        return self.__name

    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray]):  # -> Optional[MessageType]:
        if isinstance(data, cls):
            return data
        elif not isinstance(data, UInt16Data):
            data = Convert.uint16data_from_data(data=data)
        if data is None or data.value > 0x3FFF:
            # format: 00xx xxxx, xxxx xxxx
            return None
        fixed = cls.__message_types.get(data.value)
        if fixed is None:
            # name = 'MessageType-0x%04X' % ui16.value
            # fixed = cls(data=ui16, value=ui16.value, name=name)
            raise LookupError('msg type error: %d' % data.value)
        return fixed

    @classmethod
    def cache(cls, value: int, msg_type):
        cls.__message_types[value] = msg_type

    __message_types = {}  # int -> MessageType

    # fixed types for a STUN message
    BIND_REQUEST = None
    BIND_RESPONSE = None
    BIND_ERROR_RESPONSE = None
    SHARED_SECRET_REQUEST = None
    SHARED_SECRET_RESPONSE = None
    SHARED_SECRET_ERROR_RESPONSE = None


def create_type(value: int, name: str) -> MessageType:
    data = Convert.uint16data_from_value(value=value)
    fixed = MessageType(data=data, value=value, endian=data.endian, name=name)
    MessageType.cache(value=value, msg_type=fixed)
    return fixed


MessageType.BIND_REQUEST = create_type(value=0x0001, name='Binding Request')
MessageType.BIND_RESPONSE = create_type(value=0x0101, name='Binding Response')
MessageType.BIND_ERROR_RESPONSE = create_type(value=0x0111, name='Binding Error Response')
MessageType.SHARED_SECRET_REQUEST = create_type(value=0x0002, name='Shared Secret Request')
MessageType.SHARED_SECRET_RESPONSE = create_type(value=0x0102, name='Shared Secret Response')
MessageType.SHARED_SECRET_ERROR_RESPONSE = create_type(value=0x0112, name='Shared Secret Error Response')


class TransactionID(Data):

    # Magic Cookie
    MagicCookie = Convert.uint32data_from_value(value=0x2112A442)

    def __init__(self, data: Union[bytes, bytearray, ByteArray]):
        if isinstance(data, ByteArray):
            super().__init__(buffer=data.buffer, offset=data.offset, size=data.size)
        else:
            super().__init__(buffer=data)

    def __str__(self) -> str:
        clazz = self.__class__.__name__
        return '<%s: %s />' % (clazz, self.get_bytes())

    def __repr__(self) -> str:
        clazz = self.__class__.__name__
        return '<%s: %s />' % (clazz, self.get_bytes())

    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray]):  # -> Optional[TransactionID]:
        if isinstance(data, cls):
            return data
        if isinstance(data, ByteArray):
            if data.size < 16:
                return None
            elif data.size > 16:
                data = data.slice(start=0, end=16)
        else:
            data_size = len(data)
            if data_size < 16:
                return None
            elif data_size > 16:
                data = data[0:16]
        return cls(data=data)

    __number_lock = threading.Lock()
    __number_hi = Convert.int32_from_data(data=Data.random(size=4))
    __number_mi = Convert.int32_from_data(data=Data.random(size=4))
    __number_lo = Convert.int32_from_data(data=Data.random(size=4))

    @classmethod
    def generate(cls):  # -> TransactionID:
        with cls.__number_lock:
            if cls.__number_lo < 0xFFFFFFFF:
                cls.__number_lo += 1
            else:
                cls.__number_lo = 0
                if cls.__number_mi < 0xFFFFFFFF:
                    cls.__number_mi += 1
                else:
                    cls.__number_mi = 0
                    if cls.__number_hi < 0xFFFFFFFF:
                        cls.__number_hi += 1
                    else:
                        cls.__number_hi = 0
            hi = cls.__number_hi
            mi = cls.__number_mi
            lo = cls.__number_lo
        # convert integer values to Data
        hi = Convert.uint32data_from_value(value=hi)
        mi = Convert.uint32data_from_value(value=mi)
        lo = Convert.uint32data_from_value(value=lo)
        data = cls.MagicCookie.concat(hi).concat(mi).concat(lo)
        return cls(data=data)


"""
    STUN Message Structure
    ~~~~~~~~~~~~~~~~~~~~~~

   STUN messages are encoded in binary using network-oriented format
   (most significant byte or octet first, also commonly known as big-
   endian).  The transmission order is described in detail in Appendix B
   of RFC 791 [RFC0791].  Unless otherwise noted, numeric constants are
   in decimal (base 10).

   All STUN messages MUST start with a 20-byte header followed by zero
   or more Attributes.  The STUN header contains a STUN message type,
   magic cookie, transaction ID, and message length.
   
       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |0 0|     STUN Message Type     |         Message Length        |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         Magic Cookie                          |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                                
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                            Transaction ID (96 bits)                   
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                                                                      |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

                  Figure 2: Format of STUN Message Header

   The message length MUST contain the size, in bytes, of the message
   not including the 20-byte STUN header.  Since all STUN attributes are
   padded to a multiple of 4 bytes, the last 2 bits of this field are
   always zero.  This provides another way to distinguish STUN packets
   from packets of other protocols.
"""


class Header(Data):

    def __init__(self, data: Union[bytes, bytearray, ByteArray],
                 msg_type: MessageType, msg_length: int, trans_id: TransactionID):
        if isinstance(data, ByteArray):
            super().__init__(buffer=data.buffer, offset=data.offset, size=data.size)
        else:
            super().__init__(buffer=data)
        self.__msg_type = msg_type
        self.__msg_length = msg_length
        self.__trans_id = trans_id

    def __str__(self) -> str:
        clazz = self.__class__.__name__
        mt = self.msg_type
        ml = self.msg_length
        return '<%s: %d, "%s" msg_len=%d />' % (clazz, self.size, mt, ml)

    def __repr__(self) -> str:
        clazz = self.__class__.__name__
        mt = self.msg_type
        ml = self.msg_length
        return '<%s: %d, "%s" msg_len=%d />' % (clazz, self.size, mt, ml)

    @property
    def msg_type(self) -> MessageType:
        return self.__msg_type

    @property
    def msg_length(self) -> int:
        return self.__msg_length

    @property
    def trans_id(self) -> TransactionID:
        return self.__trans_id

    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray]):  # -> Optional[Header]:
        if isinstance(data, cls):
            return data
        elif not isinstance(data, ByteArray):
            data = Data(buffer=data)
        # get message type
        msg_type = MessageType.parse(data=data)
        if msg_type is None:
            return None
        pos = msg_type.size
        # get message length
        msg_len = Convert.uint16data_from_data(data=data, start=pos)
        if msg_len is None:
            return None
        pos += msg_len.size
        # get transaction ID
        trans_id = TransactionID.parse(data=data.slice(start=pos))
        if trans_id is None:
            return None
        pos += trans_id.size
        assert pos == 20, 'header length error: %d' % pos
        if data.size > pos:
            data = data.slice(start=0, end=pos)
        # create
        assert msg_len.value & 0x0003 == 0, 'msg length error: %d' % msg_len.value
        return cls(data=data, msg_type=msg_type, msg_length=msg_len.value, trans_id=trans_id)

    @classmethod
    def new(cls, msg_type: MessageType, msg_length: int, trans_id: Optional[TransactionID] = None):  # -> Header:
        assert msg_length & 0x0003 == 0, 'msg length error: %d' % msg_length
        msg_len = Convert.uint16data_from_value(value=msg_length)
        if trans_id is None:
            trans_id = TransactionID.generate()
        # concat(msg_type, msg_len, trans_id)
        size = msg_type.size + msg_len.size + trans_id.size
        data = MutableData(capacity=size)
        data.append(msg_type)
        data.append(msg_len)
        data.append(trans_id)
        return cls(data=data, msg_type=msg_type, msg_length=msg_length, trans_id=trans_id)



# -*- coding: utf-8 -*-
#
#   STUN: Session Traversal Utilities for NAT
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Session Traversal Utilities for NAT
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    [RFC] https://www.ietf.org/rfc/rfc5389.txt
    [RFC] https://www.ietf.org/rfc/rfc3489.txt
"""

from typing import Optional, Union, Dict

from udp.ba import ByteArray
from udp.ba import Endian, UInt16Data, Convert

from ..tlv import Tag16, TagParser
from ..tlv import Length16 as AttributeLength, LengthParser
from ..tlv import Value as AttributeValue, RawValue, ValueParser
from ..tlv import Triad, Parser as TriadParser


"""
    STUN Attributes
    ~~~~~~~~~~~~~~~

   After the STUN header are zero or more attributes.  Each attribute
   MUST be TLV encoded, with a 16-bit type, 16-bit length, and value.
   Each STUN attribute MUST end on a 32-bit boundary.  As mentioned
   above, all fields in an attribute are transmitted most significant
   bit first.

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |         Type                  |            Length             |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         Value (variable)                ....
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

                    Figure 4: Format of STUN Attributes
"""


#
#  Attribute Type
#
class AttributeType(Tag16):

    def __init__(self, data: Union[bytes, bytearray, ByteArray], value: int, endian: Endian, name: str):
        super().__init__(data=data, value=value, endian=endian)
        self.__name = name

    def __str__(self):
        return self.__name

    def __repr__(self):
        return self.__name

    @property
    def name(self) -> str:
        return self.__name

    @classmethod
    def parse(cls, data: Union[bytes, bytearray, ByteArray]):  # -> Optional[AttributeType]:
        if isinstance(data, cls):
            return data
        elif not isinstance(data, UInt16Data):
            data = Convert.uint16data_from_data(data=data)
        if data is not None:
            t = cls.__attribute_types.get(data.value)
            if t is None:
                name = 'Attribute-0x%+04X' % data.value
                return create_type(value=data.value, name=name)
            else:
                return t

    @classmethod
    def cache(cls, value: int, attribute_type):
        cls.__attribute_types[value] = attribute_type

    # Attribute Types in STUN message
    __attribute_types = {}  # int -> AttributeType

    # Comprehension-required range (0x0000-0x7FFF)
    # Comprehension-optional range (0x8000-0xFFFF)

    # [RFC-3489]
    MAPPED_ADDRESS = None
    RESPONSE_ADDRESS = None
    CHANGE_REQUEST = None
    SOURCE_ADDRESS = None
    CHANGED_ADDRESS = None
    USERNAME = None
    PASSWORD = None
    MESSAGE_INTEGRITY = None
    ERROR_CODE = None
    UNKNOWN_ATTRIBUTES = None
    REFLECTED_FROM = None

    # [RFC-5389]
    REALM = None
    NONCE = None
    XOR_MAPPED_ADDRESS = None

    XOR_MAPPED_ADDRESS2 = None
    XOR_ONLY = None
    SOFTWARE = None
    ALTERNATE_SERVER = None
    FINGERPRINT = None


def create_type(value: int, name: str) -> AttributeType:
    data = Convert.uint16data_from_value(value=value)
    fixed = AttributeType(data=data, value=data.value, endian=data.endian, name=name)
    AttributeType.cache(value=value, attribute_type=fixed)
    return fixed


# Comprehension-required range (0x0000-0x7FFF)
# Comprehension-optional range (0x8000-0xFFFF)

# [RFC-3489]
AttributeType.MAPPED_ADDRESS = create_type(value=0x0001, name='MAPPED-ADDRESS')
AttributeType.RESPONSE_ADDRESS = create_type(value=0x0002, name='RESPONSE-ADDRESS')
AttributeType.CHANGE_REQUEST = create_type(value=0x0003, name='CHANGE-REQUEST')
AttributeType.SOURCE_ADDRESS = create_type(value=0x0004, name='SOURCE-ADDRESS')
AttributeType.CHANGED_ADDRESS = create_type(value=0x0005, name='CHANGED-ADDRESS')
AttributeType.USERNAME = create_type(value=0x0006, name='USERNAME')
AttributeType.PASSWORD = create_type(value=0x0007, name='PASSWORD')
AttributeType.MESSAGE_INTEGRITY = create_type(value=0x0008, name='MESSAGE-INTEGRITY')
AttributeType.ERROR_CODE = create_type(value=0x0009, name='ERROR-CODE')
AttributeType.UNKNOWN_ATTRIBUTES = create_type(value=0x000A, name='UNKNOWN-ATTRIBUTES')
AttributeType.REFLECTED_FROM = create_type(value=0x000B, name='REFLECTED-FROM')

# [RFC-5389]
AttributeType.REALM = create_type(value=0x0014, name='REALM')
AttributeType.NONCE = create_type(value=0x0015, name='NONCE')
AttributeType.XOR_MAPPED_ADDRESS = create_type(value=0x0020, name='XOR-MAPPED-ADDRESS(0020)')

AttributeType.XOR_MAPPED_ADDRESS2 = create_type(value=0x8020, name='XOR-MAPPED-ADDRESS(8020)')
AttributeType.XOR_ONLY = create_type(value=0x8021, name='XOR-ONLY')
AttributeType.SOFTWARE = create_type(value=0x8022, name='SOFTWARE')
AttributeType.ALTERNATE_SERVER = create_type(value=0x8023, name='ALTERNATE-SERVER')
AttributeType.FINGERPRINT = create_type(value=0x8028, name='FINGERPRINT')


class Attribute(Triad[AttributeType, AttributeLength, AttributeValue]):

    @classmethod
    def new(cls, tag: AttributeType,
            length: Optional[AttributeLength] = None,
            value: AttributeValue = RawValue.ZERO):
        if length is None:
            length = AttributeLength.new(value=value.size)
        data = tag.concat(length).concat(value)
        return cls(data=data, tag=tag, length=length, value=value)

    # TLV parser
    parser = None

    @classmethod
    def parse_attributes(cls, data: ByteArray):  # -> List[Attribute]
        return cls.parser.parse_entries(data=data)


class AttributeParser(TriadParser[Attribute, AttributeType, AttributeLength, AttributeValue],
                      TagParser[AttributeType],
                      LengthParser[AttributeType, AttributeLength],
                      ValueParser[AttributeType, AttributeLength, AttributeValue]):
    @property
    def tag_parser(self) -> TagParser[AttributeType]:
        return self

    @property
    def length_parser(self) -> LengthParser[AttributeType, AttributeLength]:
        return self

    @property
    def value_parser(self) -> ValueParser[AttributeType, AttributeLength, AttributeValue]:
        return self

    def create_entry(self, data: ByteArray,
                     tag: AttributeType, length: AttributeLength, value: AttributeValue) -> Attribute:
        return Attribute(data=data, tag=tag, length=length, value=value)

    # TagParser
    def parse_tag(self, data: ByteArray) -> Optional[AttributeType]:
        return AttributeType.parse(data=data)

    # LengthParser
    def parse_length(self, data: ByteArray, tag: AttributeType) -> Optional[AttributeLength]:
        return AttributeLength.parse(data=data, tag=tag)

    # ValueParser
    def parse_value(self, data: ByteArray, tag: AttributeType, length: AttributeLength) -> Optional[AttributeValue]:
        parser = self.__value_parsers.get(tag.name)
        if parser is None:
            return RawValue.parse(data=data, tag=tag, length=length)
        else:
            return parser.parse_value(data=data, tag=tag, length=length)

    __value_parsers: Dict[str, ValueParser] = {}  # type => parser

    @classmethod
    def register(cls, tag: AttributeType, parser: ValueParser):
        if parser is None:
            cls.__value_parsers.pop(tag.name, None)
        else:
            cls.__value_parsers[tag.name] = parser


Attribute.parser = AttributeParser()



# -*- coding: utf-8 -*-
#
#   TLV: Tag Length Value
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import abstractmethod
from typing import Generic, Union, Optional, List

from udp.ba import ByteArray, Data

from .tag import T, TagParser
from .length import L, LengthParser
from .value import V, ValueParser
from .entry import E, Entry, EntryParser


class Parser(EntryParser, Generic[E, T, L, V]):
    """ TLV Parser """

    @property
    def tag_parser(self) -> TagParser[T]:
        """ get Tag Parser """
        raise NotImplemented

    @property
    def length_parser(self) -> LengthParser[T, L]:
        """ get Length Parser """
        raise NotImplemented

    @property
    def value_parser(self) -> ValueParser[T, L, V]:
        """ get Value Parser """
        raise NotImplemented

    @abstractmethod
    def create_entry(self, data: ByteArray, tag: T, length: L, value: V) -> E:
        """ Create TLV triad """
        raise NotImplemented

    def parse_entry(self, data: ByteArray) -> Optional[E]:
        # 1. get Tag field
        tag = self.tag_parser.parse_tag(data=data)
        if tag is None:
            return None
        offset = tag.size
        # 2. get Length field
        length = self.length_parser.parse_length(data=data.slice(start=offset), tag=tag)
        if length is None:
            # if length not defined, use the rest data as value
            value_length = data.size - offset
        else:
            value_length = length.value
            offset += length.size
        # 3. get value field
        end = offset + value_length
        value = self.value_parser.parse_value(data=data.slice(start=offset, end=end), tag=tag, length=length)
        # OK
        data = data.slice(start=0, end=end)
        return self.create_entry(data=data, tag=tag, length=length, value=value)

    def parse_entries(self, data: ByteArray) -> List[E]:
        entries = []
        while data.size > 0:
            entry = self.parse_entry(data=data)
            if entry is None:
                # data error
                break
            entries.append(entry)
            # next entry
            data = data.slice(start=entry.size)
        return entries


class Triad(Data, Entry, Generic[T, L, V]):
    """ TLV Entry """

    def __init__(self, data: Union[bytes, bytearray, ByteArray], tag: T, length: L, value: V):
        if isinstance(data, ByteArray):
            super().__init__(buffer=data.buffer, offset=data.offset, size=data.size)
        else:
            super().__init__(buffer=data)
        self.__tag = tag
        self.__length = length
        self.__value = value

    @property
    def tag(self) -> T:
        return self.__tag

    @property
    def length(self) -> L:
        return self.__length

    @property
    def value(self) -> V:
        return self.__value

    def __str__(self) -> str:
        clazz = self.__class__.__name__
        return '/* %s */ %s: "%s"' % (clazz, self.tag, self.value)

    def __repr__(self) -> str:
        clazz = self.__class__.__name__
        return '/* %s */ %s: "%s"' % (clazz, self.tag, self.value)
